/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which included in this distribution for details.
*/
#include "ZPath.h"

#include <string.h>

// defines the linear buffer increase value
#define ZBLOCKSIZE 	256

///
enum
{
	Z_CMD_NOP		= 0,
	Z_CMD_MOVETO 	= 1,
	Z_CMD_LINETO 	= 2,
	Z_CMD_QCURVETO 	= 3,
	Z_CMD_CCURVETO 	= 4,
	Z_CMD_ARCTO		= 5,
	Z_CMD_CLOSE		= 6,
	Z_STREAMID		= ('Z'<<24) | ('P'<<16) | ('T'<<8) | 'H'
};

///
ZPath::ZPath():
	_data(0),
	_used_size(0),
	_allc_size(0)
{
}

///
ZPath::ZPath(const ZPath &from):
	_data(0),
	_used_size(0),
	_allc_size(0)
{
	operator=(from);
}

///
ZPath::~ZPath()
{
	delete [] _data;
}

///
sint32 ZPath::StreamByteSize()
{
	return 2*sizeof(uint32)+_used_size*sizeof(uint32);
}

///
void *ZPath::ToStream(void *stream)
{
	uint32 *s = (uint32 *)stream;
	*s++ = Z_STREAMID;
	*s++ = _used_size;
	for(sint32 c=0;c<_used_size;c++)
		*s++ = _data[c];
	return s;
}

///
void *ZPath::FromStream(void *stream)
{
	uint32 *s = (uint32 *)stream;
	if(*s++ == Z_STREAMID)
	{
		delete [] _data;
		_data = 0;
		_allc_size = _used_size = 0;

		uint32 used_size = *s++;
		if(used_size)
		{
			_allc_size = _used_size = used_size;
			_data = new uint32 [_used_size];
			for(sint32 c=0;c<_used_size;c++)
				_data[c] = *s++;
		}
		return s;
	}
	return stream;
}

///
void ZPath::MakeEmpty()
{
	delete [] _data;
	_data = 0;
	_used_size = 0;
	_allc_size = 0;
}

///
void ZPath::AddMoveTo(ZPoint point)
{
	ExtendIfNeeded(3);
	uint32 *dst = &_data[_used_size];
	*dst++ = Z_CMD_MOVETO;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
	_used_size += 3;
}

///
void ZPath::AddLineTo(ZPoint point)
{
	ExtendIfNeeded(3);
	uint32 *dst = &_data[_used_size];
	*dst++ = Z_CMD_LINETO;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
	_used_size += 3;
}

///
void ZPath::AddCurveTo(ZPoint control, ZPoint point)
{
	ExtendIfNeeded(5);
	uint32 *dst = &_data[_used_size];
	*dst++ = Z_CMD_QCURVETO;
	*((float32 *)dst++) = (float32)control.x;
	*((float32 *)dst++) = (float32)control.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
	_used_size += 5;
}

///
void ZPath::AddCurveTo(ZPoint control1, ZPoint control2, ZPoint point)
{
	ExtendIfNeeded(7);
	uint32 *dst = &_data[_used_size];
	*dst++ = Z_CMD_CCURVETO;
	*((float32 *)dst++) = (float32)control1.x;
	*((float32 *)dst++) = (float32)control1.y;
	*((float32 *)dst++) = (float32)control2.x;
	*((float32 *)dst++) = (float32)control2.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
	_used_size += 7;
}

///
void ZPath::AddArcTo(ZPoint radius, ZPoint point, float32 rotation, sint32 sweep_arc)
{
	ExtendIfNeeded(7);
	uint32 *dst = &_data[_used_size];
	*dst++ = Z_CMD_ARCTO;
	*((float32 *)dst++) = (float32)radius.x;
	*((float32 *)dst++) = (float32)radius.y;
	*((float32 *)dst++) = (float32)point.x;
	*((float32 *)dst++) = (float32)point.y;
	*((float32 *)dst++) = (float32)rotation;
	*((sint32 *)dst++) = (sint32)sweep_arc;
	_used_size += 7;
}

///
void ZPath::AddClose()
{
	ExtendIfNeeded(1);
	uint32 *dst = &_data[_used_size];
	*dst++ = Z_CMD_CLOSE;
	_used_size += 1;
}

///
void ZPath::Iterate(Iterator *iterator) const
{
	if(iterator)
	{
		ZPoint p0,p1,p2;
		float32 rotation;
		sint32 sweep_arc;
		iterator->Init();
		for(uint32 *iter=_data;iter<(_data+_used_size);)
		{
			switch(*iter++)
			{
				case	Z_CMD_NOP:
						break;
				case	Z_CMD_MOVETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						iterator->MoveTo(p0);
						break;
				case	Z_CMD_LINETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						iterator->LineTo(p0);
						break;
				case	Z_CMD_QCURVETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						p1.x = *((float32 *)iter++);
						p1.y = *((float32 *)iter++);
						iterator->CurveTo(p0,p1);
						break;
				case	Z_CMD_CCURVETO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						p1.x = *((float32 *)iter++);
						p1.y = *((float32 *)iter++);
						p2.x = *((float32 *)iter++);
						p2.y = *((float32 *)iter++);
						iterator->CurveTo(p0,p1,p2);
						break;
				case	Z_CMD_ARCTO:
						p0.x = *((float32 *)iter++);
						p0.y = *((float32 *)iter++);
						p1.x = *((float32 *)iter++);
						p1.y = *((float32 *)iter++);
						rotation = *((float32 *)iter++);
						sweep_arc = *((sint32 *)iter++);
						iterator->ArcTo(p0,p1,rotation,sweep_arc);
				case	Z_CMD_CLOSE:
						iterator->Close();
						break;
				default:
						return;
			}
		}
	}
}

void ZPath::ExtendIfNeeded(sint32 addsize)
{
	bool realloc = false;
	for(;(_used_size+addsize)>_allc_size;)
	{
		_allc_size += ZBLOCKSIZE;
		realloc = true;
	}
	if(realloc)
	{
		uint32 *newdata = new uint32 [_allc_size];
		if(_data)
		{
			memcpy(newdata,_data,_used_size*sizeof(uint32));
			delete [] _data;
		}
		_data = newdata;
	}
}


ZPath &ZPath::operator+=(const ZPath &from)
{
	ExtendIfNeeded(from._used_size);
	uint32 *dst = &_data[_used_size];
	memcpy(dst,from._data,from._used_size*sizeof(uint32));
	_used_size += from._used_size;
	return *this;
}

ZPath &ZPath::operator=(const ZPath &from)
{
	delete [] _data;
	_data = 0;
	_allc_size = _used_size = 0;
	operator+=(from);
	return *this;
}
