/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which included in this distribution for details.
*/
#include "ZRenderer.h"
#include "ZTools.h"
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

// Quality settings
#define ZRENDERER_FRAC_Y     	     4 // 2^x levels of transparency
#define ZRENDERER_FRAC_X     		 4 // 2^x levels of transparency
#define ZRENDERER_REAL_X 			12 // real part of x values
#define ZRENDERER_REAL_Y    		11 // real part of y values
#define ZBUFFERSIZE				0xFFFFF // max size of the scanline buffer

// useful defines
#define ZRENDERER_REAL_X_MASK 	((1<<ZRENDERER_REAL_X)-1)
#define ZRENDERER_REAL_Y_MASK 	((1<<ZRENDERER_REAL_Y)-1)
#define ZRENDERER_FRAC_X_FACTOR  (1<<ZRENDERER_FRAC_X)
#define ZRENDERER_FRAC_X_MASK 	((1<<ZRENDERER_FRAC_X)-1)

/** Constructor will allocate neccessary buffers.
	Two buffers are allocated for the scanline information 
	and the radix sort. Currently they are both 1MByte.
*/
ZRenderer::ZRenderer():
	_scan(0),
	_bits(0),
	_pprw(0),
	_wdth(0),
	_hght(0),
	_mtrx(0),
	_mode(MODE_XOR)
{
	// FIXME!
	_scanbuf 	 = new uint32 [ZBUFFERSIZE];
	_scanbuf_tmp = new uint32 [ZBUFFERSIZE];
}

/** Destructor frees allocated buffers.
*/
ZRenderer::~ZRenderer()
{
	delete [] _scanbuf;
	delete [] _scanbuf_tmp;
}

/** Specify a bitmap buffer to draw to.
	This function expects a bitmap in ARGB 32 bit format, e.g.
	each pixel is 4 bytes wide. pprw defines pixel per row in 
	case this might differ from the real width of the bitmap.
	If this bitmap should be empty call ClearBuffer() to
	make sure that the alpha channel is cleared correctly.
*/
void ZRenderer::SetBuffer(uint32 *bits, uint32 wdth, uint32 hght, uint32 pprw)
{
	_bits = bits;
	_wdth = wdth;
	_hght = hght;
	_pprw = pprw;
}

/** Clears bitmap.
	This will prepare the bitmap buffer for drawing.
	Make sure that SetBuffer() was called before.	
*/
void ZRenderer::ClearBuffer()
{
	memset(_bits,0,_pprw*_hght*sizeof(uint32));
}

/** Makes bitmap drawable for standard displays.
	This method will in fact convert the buffer into
	non premultiplied alpha color space and at the
	same time apply a background color to the current
	image. This is neccessary to draw a bitmap to
	standard displays which do not support premultiplied
	alpha color space drawing. This is the case for
	all operating systems known.
*/
void ZRenderer::FinalizeBuffer(uint32 color)
{
	color |= 0xFF000000;
	uint32 *bits = _bits;
	for(sint32 y=0;y<_hght;y++)
	{
		for(sint32 x=0;x<_wdth;x++)
		{
			uint32 s = bits[x];
			uint32 sa = s>>24;
			if( sa != 0x00 )
			{
				if( sa != 0xFF )
				{
					uint32 sr = (0xFF*ZMIN(sa,((s>>16)&0xFF))+(0xFF-sa)*((color>>16)&0xFF)); sr = (sr+(sr>>8)+1)>>8;
					uint32 sg = (0xFF*ZMIN(sa,((s>> 8)&0xFF))+(0xFF-sa)*((color>> 8)&0xFF)); sg = (sg+(sg>>8)+1)>>8;
					uint32 sb = (0xFF*ZMIN(sa,((s    )&0xFF))+(0xFF-sa)*((color    )&0xFF)); sb = (sb+(sb>>8)+1)>>8;
					bits[x] = 0xFF000000|(sr<<16)|(sg<< 8)|(sb);
				}
			}
			else
			{
				bits[x] = color;
			}
		}
		bits += _pprw;
	}
}

/**	Draw the specified path to the bitmap.
	The stencil defines the filling to be used. mode is
	either MODE_XOR for an odd/even filling rule or 
	MODE_ZERO for zero/nonzero filling rule.
*/
void ZRenderer::DrawPath(const ZPath &path, ZStencil &stencil, sint32 mode)
{
	_scan = _scanbuf;

	_xmin = ZMAXFLOAT32;
	_xmax = ZMINFLOAT32;
	_ymin = ZMAXFLOAT32;
	_ymax = ZMINFLOAT32;
	_mode = mode;

	// 1. Split path into segments
	// 2. Create scanline information
	path.Iterate(this);

	// 3. Sort scanline information
	radix_sort(_scanbuf,_scanbuf_tmp,_scan-_scanbuf);

	// 4. Set up stencil
	_stnc = &stencil;
	if(_mtrx) _stnc->SetPathMatrix(*_mtrx);
	_stnc->SetBounds(_xmin,_ymin,_xmax,_ymax);
	_mono = _stnc->IsMonotone();

	// 5. Draw scanline information
	DrawBuffer();

	_stnc = 0;
}


/** Draw the specified path to the bitmap.
    The stencil defines the filling to be used. mode is
    either MODE_XOR for an odd/even filling rule or
    MODE_ZERO for zero/nonzero filling rule. In addition
	you can specify a matrix which will be applied after
	the curves were sampled.
*/
void ZRenderer::DrawPath(const ZPath &path, const ZMatrix &matrix, ZStencil &stencil, sint32 mode)
{
	_mtrx = &matrix;
	DrawPath(path,stencil, mode);
	_mtrx = 0;
}


//
void ZRenderer::MergePixels(sint32 x, sint32 y, sint32 count, uint08 opacity)
{
	// Is color monotone or f.ex. a gradient?
	if(_mono)
	{
		uint32 *dst = &_bits[y*_pprw + x];
		// Get Monotone color
		uint32 s = _stnc->GetPixel(false);
		uint32 sa = (s>>24);
		// Apply opacity if necessary
		if(opacity!=0xFF)
		{
			uint32 Ca = ((s>>24)&0xFF)*(uint32)opacity;
			uint32 Cr = ((s>>16)&0xFF)*(uint32)opacity;
			uint32 Cg = ((s>> 8)&0xFF)*(uint32)opacity;
			uint32 Cb = ((s    )&0xFF)*(uint32)opacity;
			sa = (((Ca+(Ca>>8)+1)>>8));
			s  = (sa<<24)| 
				 (((Cr+(Cr>>8)+1)>>8)<<16)|
				 (((Cg+(Cg>>8)+1)>>8)<< 8)|
				 (((Cb+(Cb>>8)+1)>>8)    );
		}
		if(sa!=0x00)
		{
			sa = 0xFF-sa;
			for(sint32 c=0;c<count;c++)
			{
				uint32 d = dst[c];
				uint32 da = d>>24;
				if(da==0x00)
				{
					dst[c] = s;
				}
				else
				{
					// Only do the real composition if neccessary
					if(da!=0xFF)
					{
						// Compose under current pixel
						da = 0xFF-da;
						uint32 Ca = (0xFF*0xFF-sa*da);					   Ca = (Ca+(Ca>>8)+1)>>8;
						uint32 Cr = ((s>>16)&0xFF)*da+((d>>16)&0xFF)*0xFF; Cr = (Cr+(Cr>>8)+1)>>8;
						uint32 Cg = ((s>> 8)&0xFF)*da+((d>> 8)&0xFF)*0xFF; Cg = (Cg+(Cg>>8)+1)>>8;
						uint32 Cb = ((s    )&0xFF)*da+((d    )&0xFF)*0xFF; Cb = (Cb+(Cb>>8)+1)>>8;
						dst[c] = (Ca<<24)|(Cr<<16)|(Cg<<8)|Cb;
					}
				}
			}
		}
	}
	else
	{
		uint32 *dst = &_bits[y*_pprw + x];
		// Check if we are already on good position 
		if(_opsx != x || _opsy != y)
		{
			_stnc->SetPosition(x,y);
		}
		for(sint32 c=0;c<count;c++)
		{
			uint32 d = dst[c];
			uint32 da = d>>24;
			if(da!=0x00)
			{
				if(da!=0xFF)
				{
					// Obtain complex pixel
					uint32 s  = _stnc->GetPixel(false);
					uint32 sa = (s>>24);
					if(sa!=0x00)
					{
						// Apply opacity only when necessary
						if(opacity==0xFF)
						{
							// Compose under current pixel
							sa = 0xFF-sa;
							da = 0xFF-da;
							uint32 Ca = (0xFF*0xFF-sa*da);					   Ca = (Ca+(Ca>>8)+1)>>8;
							uint32 Cr = ((s>>16)&0xFF)*da+((d>>16)&0xFF)*0xFF; Cr = (Cr+(Cr>>8)+1)>>8;
							uint32 Cg = ((s>> 8)&0xFF)*da+((d>> 8)&0xFF)*0xFF; Cg = (Cg+(Cg>>8)+1)>>8;
							uint32 Cb = ((s    )&0xFF)*da+((d    )&0xFF)*0xFF; Cb = (Cb+(Cb>>8)+1)>>8;
							dst[c] = (Ca<<24)|(Cr<<16)|(Cg<<8)|Cb;
						}
						else
						{
							// Compose under current pixel
							sa *= opacity; sa = (sa+(sa>>8)+1)>>8;
							sa = 0xFF-sa;
							da = 0xFF-da;
							uint32 Ca = (0xFF*0xFF-sa*da);					   Ca = (Ca+(Ca>>8)+1)>>8;
							da *= opacity; da = (da+(da>>8)+1)>>8;
							uint32 Cr = ((s>>16)&0xFF)*da+((d>>16)&0xFF)*0xFF; Cr = (Cr+(Cr>>8)+1)>>8;
							uint32 Cg = ((s>> 8)&0xFF)*da+((d>> 8)&0xFF)*0xFF; Cg = (Cg+(Cg>>8)+1)>>8;
							uint32 Cb = ((s    )&0xFF)*da+((d    )&0xFF)*0xFF; Cb = (Cb+(Cb>>8)+1)>>8;
							dst[c] = (Ca<<24)|(Cr<<16)|(Cg<<8)|Cb;
						}
					}
				}
				else
				{
					// We do not need the color, so tell the stencil to skip this pixel
					_stnc->GetPixel(true);
				}
			}
			else
			{
				// Apply opacity only if necessary
				if(opacity == 0xFF)
				{
					// Obtain complex pixel and copy it
					dst[c] = _stnc->GetPixel(false);
				}
				else
				{
					uint32 s  = _stnc->GetPixel(false);
					uint32 Ca = ((s>>24)&0xFF)*opacity;
					uint32 Cr = ((s>>16)&0xFF)*opacity;
					uint32 Cg = ((s>> 8)&0xFF)*opacity;
					uint32 Cb = ((s    )&0xFF)*opacity;
					dst[c] =(((Ca+(Ca>>8)+1)>>8)<<24)|
							(((Cr+(Cr>>8)+1)>>8)<<16)|
							(((Cg+(Cg>>8)+1)>>8)<< 8)|
							(((Cb+(Cb>>8)+1)>>8)    );
				}
			}
		}

		// Set current position 
		_opsx = x+count;
		_opsy = y; 
	}
}

//
void ZRenderer::DrawBuffer()
{
 	sint32 curd=0;	// vertical opacity
	sint32 cure=0;  // vertical opacity backup
	sint32 cura=0;	// right opacity sum
	sint32 cula=0;	// left opacity sum

	uint32 cury=0;	// current vertical position
	uint32 curx=0;	// current horizontal position
	uint32 olds=0;

	sint32 c   =0;
	uint32 curs=0;

	// Draw scanline information we got for this path
	sint32 count = _scan-_scanbuf;

	_opsx = -1;
	_opsy = -1; 

	for(;c<count;c++)
	{
		// make sure we end the last scanline
		curs = _scanbuf[c];

dolastscanline:

		// Setup new vertical and horizontal values (non fixed integer)
		uint32 newy = ((curs>>(ZRENDERER_REAL_X+ZRENDERER_FRAC_X+1))&ZRENDERER_REAL_Y_MASK);
		uint32 newx = ((curs>>(                 ZRENDERER_FRAC_X+1))&ZRENDERER_REAL_X_MASK);

		// Check if we are still on the same line or column
		if((newx != curx) || (newy != cury))
		{
			// Compute opacity we got so far for this pixel
			sint32 alp = (256*cure)/(ZRENDERER_FRAC_Y                            )+
			             (256*cula)/(ZRENDERER_FRAC_Y*(ZRENDERER_FRAC_X_FACTOR-1))+ 
			             (256*cura)/(ZRENDERER_FRAC_Y*(ZRENDERER_FRAC_X_FACTOR-1));
			if(alp)
			{
				if(_mode == MODE_XOR)
					alp = (alp&0x100) ? (0xFF-(alp&0xFF)) : (alp&0xFF);
				else 
					alp = ZMIN(255,abs(alp));
				if(alp) MergePixels(curx,cury,1,(uint08)alp);
			}
			cure = curd;

			// If we are still on the same line, we fill the line
			// until the new column using the current opacity
			if(newy == cury)
			{
				// Optimize
				if(curd)
				{
					sint32 alp = (256*curd)/ZRENDERER_FRAC_Y;
					if(alp)
					{
						if(_mode == MODE_XOR)
							alp = (alp&0x100) ? (0xFF-(alp&0xFF)) : (alp&0xFF);
						else
							alp = ZMIN(255,abs(alp));
						if(alp) MergePixels(curx+1,cury,newx-curx-1,(uint08)alp);
					}
				}
			}
			else
			{
				// If we are on a new line, reset current opacity
				cury = newy;
				curd = cure = 0;
			}

			// Reset values for next pixel(s)
			curx = newx;
			cura = cula = 0;
		}

		// Check direction and update horizontal opacity values accordingly
		// and also the vertical opacity (curd)
		if(curs&1)
		{
			curd++; 
			cula += ((~(curs>>1))&ZRENDERER_FRAC_X_MASK);
		}
		else
		{
			curd--; 
			cura -= ((~(curs>>1))&ZRENDERER_FRAC_X_MASK);
		}
	}

	if(c == count)
	{
		curs = 0;
		goto dolastscanline;
	}
}

//
void ZRenderer::DoScanline(float32 Sx, float32 Sy, float32 Ex, float32 Ey)
{
	sint32 xmin;
	sint32 xmax;
	sint32 ymin;
	sint32 ymax;
	sint32 incx;
	sint32 x;
	sint32 dire;

	// determine direction and bounds
	if(Sx < Ex)
	{
		xmin = (sint32)floor(Sx*ZRENDERER_FRAC_X_FACTOR);
		xmax = (sint32)floor(Ex*ZRENDERER_FRAC_X_FACTOR);
		if(Sy < Ey)
		{
			incx = +1;
			x	 = xmin;
			ymin = (sint32)floor(Sy*(float32)ZRENDERER_FRAC_Y);
			ymax = (sint32)floor(Ey*(float32)ZRENDERER_FRAC_Y);
			dire = 1;
		}
		else
		{
			incx = -1;
			x	 = xmax;
			ymin = (sint32)floor(Ey*(float32)ZRENDERER_FRAC_Y);
			ymax = (sint32)floor(Sy*(float32)ZRENDERER_FRAC_Y);
			dire = 0;
		}
	}
	else
	{
		xmin = (sint32)floor(Ex*ZRENDERER_FRAC_X_FACTOR);
		xmax = (sint32)floor(Sx*ZRENDERER_FRAC_X_FACTOR);
		if(Sy < Ey)
		{
			incx = -1;
			x	 = xmax;
			ymin = (sint32)floor(Sy*(float32)ZRENDERER_FRAC_Y);
			ymax = (sint32)floor(Ey*(float32)ZRENDERER_FRAC_Y);
			dire = 1;
		}
		else
		{
			incx = +1;
			x	 = xmin;
			ymin = (sint32)floor(Ey*(float32)ZRENDERER_FRAC_Y);
			ymax = (sint32)floor(Sy*(float32)ZRENDERER_FRAC_Y);
			dire = 0;
		}
	}

	// don't process scanlines which are out of bounds
	if(
		((ymin<0                         ) && (ymax<0                         )) ||
		((ymin>=(_hght*ZRENDERER_FRAC_Y )) && (ymax>=(_hght*ZRENDERER_FRAC_Y )))
	) return;

	// determine steepness
	sint32 n = abs(xmax-xmin);
	sint32 d = abs(ymax-ymin);
	sint32 i = d;

	// clip bottom
	ymax = ZMIN(ymax,_hght*ZRENDERER_FRAC_Y);

	// fill scanline buffer
	for(sint32 y=ymin; y<ymax; y++)
	{
		// This if() here could be avoided with
		// a precomputation. It would also
		// avoid useless interations...
		if(y>=0)
		{
			*_scan++ = ((y/ZRENDERER_FRAC_Y)<<(ZRENDERER_REAL_X+ZRENDERER_FRAC_X+1)) | 
						(ZMAX(0,ZMIN((_wdth*ZRENDERER_FRAC_X_FACTOR)-1,x))<<1) | 
						dire;
			if((_scan-_scanbuf) > ZBUFFERSIZE) return;
		}
		i += n;
		// go to next scanline
		if(i>d)
		{
			sint32 idivd = (i-1)/d;
			x+= incx*idivd;
			i-= d   *idivd;
		}
	}
}


void ZRenderer::LineTo(ZPoint point)
{
	// All new entry points first get transformed
	ZPoint a(CurrentPoint());
	ZPoint b(point);

	// This is used to fill with a gradient/pattern later
	// it computes the non transformed bounds of the 
	// current path
	_ymin = ZMIN(_ymin,a.y);
	_ymax = ZMAX(_ymax,b.y);
	_xmin = ZMIN(_xmin,a.x);
	_xmax = ZMAX(_xmax,b.x);

	if(_mtrx)
	{
		a *= *_mtrx;
		b *= *_mtrx;
	}

	DoScanline(a.x,a.y,b.x,b.y);

	// Make sure that the Sketcher updates its information
	ZSketcher::LineTo(point);
}
