/*
	Copyright (c) 2000, Tinic Uro <tinic@tinic.net> 
	All rights reserved. Please read the LICENSE file 
	which included in this distribution for details.
*/
#include "ZSketcher.h"
#include <math.h>

#define ZSUBDIVIDE 				24 // subdivide curves by this number of steps
#define ZSUBDIVIDESTEP 			(1.f/ZSUBDIVIDE)
#define ZSUBDIVIDESTEP2 		(ZSUBDIVIDESTEP*ZSUBDIVIDESTEP)
#define ZSUBDIVIDESTEP3 		(ZSUBDIVIDESTEP*ZSUBDIVIDESTEP*ZSUBDIVIDESTEP)

/** A new path is about to be sampled.
	Implement this virtual method when you need to initialize
	things before a new path is sampled.
*/
void ZSketcher::Init()
{
	_strt = true;
}

/**	Add a line to the output path.
	When inhereting implement this virtual method to obtain the 
	sampled curves. In fact sampling is done by subdividing the curves 
	into linear segments. The sampling methods simply call LineTo() with
	their sampled values. Be sure to call ZSketcher::LineTo() somewhere in 
	your methid or otherwise you might experience interesting results. 
	Although this method does almost nothing really productive in the default 
	implementation it still needs to be called to update the cursor position.
*/
void ZSketcher::LineTo(ZPoint point)
{
	_curr_point = point;
}

void ZSketcher::MoveTo(ZPoint point)
{
	if(_strt)
	{
		_strt_point = point;
		_strt = false;
	}
	_curr_point = point;
}


void ZSketcher::CurveTo(ZPoint control, ZPoint point)
{
	// Compute forward difference values for a quadratic
	// curve of type A*(1-t)^2 + 2*B*t*(1-t) + C*t^2

	ZPoint f(_curr_point);
	ZPoint tmp((_curr_point-control*2.+point)*(float32)ZSUBDIVIDESTEP2);
	ZPoint ddf(tmp*2.);
	ZPoint df(tmp+(control-_curr_point)*(float32)(2.*ZSUBDIVIDESTEP));

	for(sint32 c=0;c<(ZSUBDIVIDE-1);c++)
	{
		f += df; df += ddf; LineTo(f);
	}

	// We specify the last point manually since
	// we obtain rounding errors during the
	// forward difference computation.
	LineTo(point);
}

void ZSketcher::CurveTo(ZPoint control1, ZPoint control2, ZPoint point)
{
	float32 pre1 = 3*ZSUBDIVIDESTEP;
	float32 pre2 = 3*ZSUBDIVIDESTEP2;
	float32 pre3 = ZSUBDIVIDESTEP3;
	float32 pre4 = 6*ZSUBDIVIDESTEP2;
	float32 pre5 = 6*ZSUBDIVIDESTEP3;
	
	ZPoint tmp1(_curr_point-control1*2+control2);
	ZPoint tmp2((control1-control2)*3-_curr_point+point);
	
	ZPoint f(_curr_point);
	ZPoint df((control1-_curr_point)*pre1+tmp1*pre2+tmp2*pre3);
	ZPoint ddf(tmp1*pre4+tmp2*pre5);
	ZPoint dddf(tmp2*pre5);
	
	for(sint32 c=0;c<(ZSUBDIVIDE-1);c++)
	{
		f += df; df += ddf; ddf += dddf; LineTo(f);
	}

	LineTo(point);
}

void ZSketcher::ArcTo(ZPoint radius, ZPoint point, float32 rotation, sint32 sweep_arc)
{
	// FINISHME
	ZPoint Center(_curr_point-radius);

	float32 start	= (float32)acos((_curr_point.x-Center.x)/radius.x);
	float32 end		= (float32)acos((point.x-Center.x)/radius.x);
	float32 step 	= (2*ZPI/(ZSUBDIVIDE*4));

	if(sweep_arc)
	{
		if(end < start) end += 2*ZPI;
		for(float32 r=(start+step);r<(end-step);r+=step)
		{
			LineTo(ZPoint(
				Center.x+radius.x*(float32)cos(r),
				Center.y+radius.y*(float32)sin(r))
			);
		}
		LineTo(point);
	}
	else
	{
		if(end > start) end -= 2*ZPI;
		for(float32 r=(start-step);r>(end+step);r-=step)
		{
			LineTo(ZPoint(
				Center.x+radius.x*(float32)cos(r),
				Center.y+radius.y*(float32)sin(r))
			);
		}
		LineTo(point);
	}
}

void ZSketcher::Close()
{
	// Connect start point with end point
	LineTo(_strt_point);
	_strt = true;
}

/*
	Computation of the quadratic forward difference values

	Generic form:
	A(1-t)^2 + 2Bt(1-t) + C(t^2)

	Isolate t:
	= A(1+t^2-2t) + 2Bt - 2B(t^2) + C(t^2)
	= A + A(t^2) - 2At + 2Bt - 2B(t^2) + C(t^2)
	= (A - 2B + C)(t^2) + (-2A + 2B)t + A

	Substitute factors:
	a = (A - 2B + C)
	b = (-2A + 2B)
	c = A

	Generic quadratic curve:
	f(t) = a(t^2) + bt + c

    First derivation:
	f(t+dt)-f(t) = a((t+dt)^2) + b(t+dt) + c - (a(t^2) + bt + c)
	= a(t^2+dt^2+2tdt) + bt + bdt + c - a(t^2) - bt - c
	= a(t^2) + a(dt^2) + a2tdt + bdt - a(t^2)
	= a(dt^2) + a2tdt + bdt

	df(t) = a2tdt + a(dt^2) + bdt

    Second derivation:
	df(t+dt)-df(t) = a2dt(t+dt) + a(dt^2) + bdt - (a2tdt + a(dt^2) + bdt)
	= 2a(dt^2)

	ddf(t) = 2a(dt^2);

    Result:
	f(0) = A
	df(0) = (A - 2B + C)*dt^2 + 2*(-A + B)*dt
	ddf(0) = 2(A - 2B + C)*dt^2
*/
